#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>

typedef unsigned int uint;

struct sparseMatrix {

  uint size;  // Size of whole matrix
  uint nNz;   // # of non zero terms

  std::vector<uint  > rows,cols; // row and column indices of non zero terms
  std::vector<double> vals;      // non zero terms

};

// Fills the (NxN) sparse matrix with the discretization of a 1D heat equation
// dt u - laplace u = 0
// with Dirichlet BC for first and last coefficients (suits with our case)
void fill1DHeat(sparseMatrix& M, double dx, double dt, int N) {

  if (N<3) {
    std::cerr << "fill1DHeat called with too low number of points N." << std::endl;
    abort();
  }

  M.size = N;
  // N-2 lines with 3 coefficients + 2
  M.nNz  = (N-2)*3 + 2;

  M.rows.push_back(0);
  M.cols.push_back(0);
  M.vals.push_back(1.);
  for (uint i=1u; i<N-1; i++) {
    M.rows.push_back(i);
    M.cols.push_back(i-1);
    M.vals.push_back(-1./(dx*dx));
    M.rows.push_back(i);
    M.cols.push_back(i+1);
    M.vals.push_back(-1./(dx*dx));
    M.rows.push_back(i);
    M.cols.push_back(i);
    M.vals.push_back(2./(dx*dx) + 1./dt);
  }
  M.rows.push_back(N-1);
  M.cols.push_back(N-1);
  M.vals.push_back(1.);

}

// Reads coefficients of a sparse matrix from a text file
// The file contains the number of non zero terms on its first line
// then on each line the row index, column index and value of non zero term
sparseMatrix readMatrix(std::string fileName) {

  sparseMatrix M;
  std::ifstream mFile(fileName.c_str());
  
  M.size = 0;

  mFile >> M.nNz;
  M.rows.resize(M.nNz);
  M.cols.resize(M.nNz);
  M.vals.resize(M.nNz);
  
  for (uint i=0u; i<M.nNz; i++) {
    mFile >> M.rows[i] >> M.cols[i] >> M.vals[i];
    if (M.rows[i] > M.size) M.size = M.rows[i];
    if (M.cols[i] > M.size) M.size = M.cols[i];
  }

  M.size++; // Indices start at 0!
  return M;

}

// Prints on screen the contents of a sparse matrix
void displayMatrix(const sparseMatrix& M) {
  std::cout << "Sparse matrix: size: " << M.size << ", non-zero terms (" << M.nNz << "):" << std::endl;
  for (uint i=0u; i<M.nNz; i++)
    std::cout << "(" << M.rows[i] << "," << M.cols[i] << ") : " << M.vals[i] << std::endl;
}

// Computes trace of matrix (i.e. sum of diagonal terms)
double trace(const sparseMatrix& M) {
  double res = 0.e0;
  for (uint i=0u; i<M.nNz; i++)
    if (M.rows[i]==M.cols[i])
      res += M.vals[i];
  return res;
}

// Computes infinite norm of matrix (max over rows of sum of abs of coefficients)
double inftNorm(const sparseMatrix& M) {
  if (M.nNz==0) {
    std::cout << "Warning: empty matrix !" << std::endl;
    return 0.;
  }  
 
  // Compute sums
  std::vector<double> maxs(M.size,0.);
  for (uint i=0u; i<M.nNz; i++)
    maxs[M.rows[i]] += abs(M.vals[i]);

  // Get max of sums
  double res = maxs[0];
  for (uint i=1u; i<M.size; i++)
    if (maxs[i]>res) res = maxs[i];
  return res;
}

// Fills res with the matrix vector product Mx.
void matMul(const sparseMatrix& M, const std::vector<double>& x, std::vector<double>& res) {

  // Check matrix and vector shapes
  if (x.size()>M.size)
    std::cout << "matMul: Warning, will only use " << M.size << " first coefficients of x (size of matrix)." << std::endl;
  else if (x.size()<M.size) {
    std::cerr << "matMul: Vector (" << x.size() << ") is smaller than matrix (" << M.size << ")." << std::endl;
    abort();
  }

  // (re)initialize result
  res.clear();
  for (uint i=0u; i<x.size(); i++)
    res.push_back(0.);

  // Add each matrix coefficient contribution
  for (uint i=0u; i<M.nNz; i++)
    res[M.rows[i]] += M.vals[i]*x[M.cols[i]];

}

double inftNorm(const std::vector<double>& x) {
  if (x.size()==0) {
    std::cout << "Warning: empty vector !" << std::endl;
    return 0.;
  }  
  double res = abs(x[0]);
  for (uint i=1u; i<x.size(); i++)
    if (abs(x[i])>res) 
      res = abs(x[i]);
  return res;
}

// Uses power iteration method to compute dominating eigen value of sparse matrix
double powerIteration(const sparseMatrix& M, int maxIters = 100, double tolerance = 1.e-4) {

  // Initialize algorithm
  std::vector<double> bk  (M.size,0.);
  std::vector<double> Mbk (M.size,0.);
  std::vector<double> bkP1(M.size,0.);
  for (double& b: bk)
    b = std::rand()/(double)RAND_MAX;

  matMul(M,bk,Mbk);
  double normk = inftNorm(Mbk), normkP1;
  double diff  = 1.e3;
  int    iter  = 0;

  // Iterate
  do {
    // Compute bk+1 = Mbk / norm(Mbk)
    for (uint i=0u; i<M.size; i++)
      bkP1[i] = Mbk[i] / normk;    
    
    // Compute MbK+1 and its norm (we put the result in Mbk)
    matMul(M,bkP1,Mbk);
    normkP1 = inftNorm(Mbk);
    
    // Determine if we have to stop (relative difference of result)
    diff = abs(normkP1-normk)/normk;
    iter++;

    // Prepare next iteration
    normk = normkP1;


  } while (diff > tolerance and iter < maxIters); 
 
  std::cout << "Power iteration stopped after " << iter << " iterations." << std::endl; 

  return normk;
}

// Simple scalar product
double dotProd(const std::vector<double>& v1, const std::vector<double>& v2) {

  double res;
  if (v1.size() != v2.size()) {
    std::cerr << "dotProd: incompatible sizes of vectors" << std::endl;
    abort();
  }
  for(uint i=0u;i<v1.size();i++)
    res += v1[i]*v2[i];
  return res;

}

// Solves the linear system Ax=b with conjugate gradient where
// matrix is sparse and vectors dense
// Convergence is tested with absolute and relative tolerance
void solveCG(const sparseMatrix& A, const std::vector<double>& b, std::vector<double>& x) {

  // Check sizes
  if (A.size != b.size() or x.size() != A.size) {
    std::cerr << "solveCG: incompatible sizes of vectors and matrix " << std::endl;
    abort();
  }

  double relTol  = 1.e-5;
  double absTol  = 1.e-9;
  double normR   = 1.e3,normRPrev;
  double diff    = 1.e3;
  int    iterMax = 100;
  int    iter    = 0;

  std::vector<double> residual (A.size);
  std::vector<double> residual2(A.size);
  std::vector<double> tmp(A.size);
  std::vector<double> direction(A.size);

  // Init: we start from x0 = 0, so first residual is just b
  // First direction is b as well
  for (uint i=0u;i<A.size;i++) {
    x        [i] = 0.0;
    residual [i] = b[i];
    direction[i] = residual[i];
  }

  do {

    normRPrev = normR;

    matMul(A,direction,tmp);
    double distance = dotProd(residual,residual)/dotProd(direction,tmp);
    normR = 0.e0;
    for (uint i=0u;i<A.size;i++) {
      x[i]         += distance*direction[i];
      residual2[i]  = residual[i] - distance*tmp[i];
      normR        += pow(residual2[i],2);
    }
    normR = sqrt(normR);
    double distance2 = dotProd(residual2,residual2)/dotProd(residual,residual);
    for (uint i=0u;i<A.size;i++) {
      direction[i] = residual2[i] + distance2*direction[i];
      residual [i] = residual2[i];
    } 

    diff = fabs(normRPrev-normR)/normRPrev;
    iter ++;
  } while (diff > relTol and normR > absTol and iter < iterMax); 

  std::cout << "CG finished after " << iter << " iterations." << std::endl;

}


int main() {

  sparseMatrix A;
  int          N  = 100;
  double       dx = M_PI/(N-1.);
  double       dt = 0.01;
  fill1DHeat(A,dx,dt,N);

  // We test heat(u) = O for x in [0,pi]
  // with u0 = sin(x)
  // solution is exp(-t)sin(x)
  std::vector<double> B(N);
  for (uint i=0u;i<B.size();i++)
    B[i] = sin(i*dx)/dt;

  std::vector<double> u(N);
  solveCG(A,B,u);

  double normSol  = 0.e0;
  double normDiff = 0.e0;
  for (uint i=0u;i<B.size();i++) {
    normSol  += pow(exp(-dt)*sin(i*dx),2);
    normDiff += pow(exp(-dt)*sin(i*dx)-u[i],2);
  }
  std::cout << "Relative error between solution and approx by CG: " << sqrt(normDiff/normSol) << std::endl;

}



